﻿/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace EventsModel
{
	using System;
	using System.Collections.Generic;
	using System.IO;

	using MicrosoftResearch.Infer.Maths;
	using MicrosoftResearch.Infer.Utils;
    using System.Globalization;
    using System.Diagnostics;

	public class DataUtils
	{
        public static double valueToIgnore = 0;
        public static bool useVenuesInsteadOfStops = true;
        public static bool useCategoriesAsFeatures = true;
        public static bool useTopicsAsFeatures = true;

        public static Vector[][] Read(string filename, bool ignoreHeader, out Vector[] baseFeatures, out double[] targets, out string[] dates, out string[][] titles)
        {
            List<Event> xBase = new List<Event>();
            List<List<Event>> xEvents = new List<List<Event>>();
            List<double> targetsList = new List<double>();
            List<string> datesList = new List<string>();
            List<string[]> titlesList = new List<string[]>();
            string line;
            Dictionary<string, int> categoriesDict = new Dictionary<string, int>();
            int nextCategoryID = 0;
            Dictionary<string, int> stopsDict = new Dictionary<string, int>();
            int nextStopID = 0;
            Dictionary<string, int> venuesDict = new Dictionary<string, int>();
            int nextVenueID = 0;
            int numTopics = -1;

            // Find start.
            int currentLocation = 0;
            using (StreamReader reader = new StreamReader(filename))
            {
                if (ignoreHeader)
                {
                    line = reader.ReadLine();
                }

                // Now read data.
                int readItems = 0;
                while (!reader.EndOfStream)
                {
                    line = reader.ReadLine();
                    string[] pieces = line.Split('\t');
                    datesList.Add(pieces[0] + "," + pieces[1] + "," + pieces[4] + "," + pieces[5]);
                    string stop_id = pieces[0];
                    int dow = Int32.Parse(pieces[2]);
                    int holiday = 0;
                    if (pieces[3].Equals("True"))
                    {
                        holiday = 1;
                    }
                    int start_time = Int32.Parse(pieces[4], CultureInfo.InvariantCulture);
                    int end_time = Int32.Parse(pieces[5], CultureInfo.InvariantCulture);
                    double area = Double.Parse(pieces[6], CultureInfo.InvariantCulture);
                    double rmse = Double.Parse(pieces[7], CultureInfo.InvariantCulture);
                    double mae = Double.Parse(pieces[8], CultureInfo.InvariantCulture);
                    double ae = Double.Parse(pieces[9], CultureInfo.InvariantCulture);
                    double overall_mean = Double.Parse(pieces[10], CultureInfo.InvariantCulture);
                    double expected_mean = Double.Parse(pieces[11], CultureInfo.InvariantCulture);
                    double observed_mean = Double.Parse(pieces[12], CultureInfo.InvariantCulture);
                    double target = ae;
                    targetsList.Add(target);

                    Event dummy = new Event();
                    if (!stopsDict.ContainsKey(stop_id))
                    {
                        stopsDict.Add(stop_id, nextStopID);
                        nextStopID++;
                    }
                    stopsDict.TryGetValue(stop_id, out dummy.stop_id);
                    dummy.dow = dow;
                    dummy.holiday = holiday;
                    dummy.start_time = start_time;
                    dummy.end_time = end_time;
                    dummy.overall_mean = overall_mean;
                    dummy.expected_mean = expected_mean;
                    dummy.observed_mean = observed_mean;
                    xBase.Add(dummy);

                    int nEvents = pieces.Length - 13;

                    List<Event> events = new List<Event>();
                    List<string> titlesSubList = new List<string>();
                    for (int i = 0; i < nEvents; i++)
                    {
                        string[] eventFeatures = pieces[i+13].Split(',');
                        //Console.WriteLine(StringUtil.ArrayToString(eventFeatures));

                        Event e = new Event();
                        //E0-001-052768148-0,0.401171375873,False,None,0|0|0|0.17422345354|0.26210840678|0|0|0|0|0|0|0|0|0|0|0.265635237484|0|0|0|0|0.204699568853|0|0|0|0
                        stopsDict.TryGetValue(stop_id, out e.stop_id);
                        e.dow = dow;
                        e.holiday = holiday;
                        e.start_time = start_time;
                        e.end_time = end_time;
                        e.overall_mean = overall_mean;
                        e.expected_mean = expected_mean;
                        e.observed_mean = observed_mean;
                        e.id = eventFeatures[0];
                        titlesSubList.Add(eventFeatures[1]);
                        if (!venuesDict.ContainsKey(eventFeatures[2]))
                        {
                            venuesDict.Add(eventFeatures[2], nextVenueID);
                            nextVenueID++;
                        }
                        venuesDict.TryGetValue(eventFeatures[2], out e.venue);
                        e.distance = Double.Parse(eventFeatures[3], CultureInfo.InvariantCulture);
                        e.all_day = 0;
                        if (eventFeatures[4].Equals("True"))
                        {
                            e.all_day = 1;
                        }
                        e.repeated_count = 0;
                        if (!eventFeatures[5].Equals("None"))
                        {
                            e.repeated_count = Int32.Parse(eventFeatures[5]);
                        }

                        if (useCategoriesAsFeatures)
                        {
                            // process categories
                            string[] eventCategories = eventFeatures[6].Split('|');
                            int[] eventCategoriesInt = new int[eventCategories.Length];
                            Array.Clear(eventCategoriesInt, 0, eventCategoriesInt.Length);
                            for (int j = 0; j < eventCategories.Length; j++)
                            {
                                if (!categoriesDict.ContainsKey(eventCategories[j]))
                                {
                                    categoriesDict.Add(eventCategories[j], nextCategoryID);
                                    nextCategoryID++;
                                }
                                categoriesDict.TryGetValue(eventCategories[j], out eventCategoriesInt[j]);
                            }
                            e.categories = eventCategoriesInt;
                        }

                        if (useTopicsAsFeatures)
                        {
                            string[] eventTopics = eventFeatures[7].Split('|');
                            double[] eventTopicsDouble = new double[eventTopics.Length];
                            for (int j = 0; j < eventTopics.Length; j++)
                            {
                                eventTopicsDouble[j] = Double.Parse(eventTopics[j], CultureInfo.InvariantCulture);
                            }
                            e.topics = eventTopicsDouble;
                            if (numTopics == -1)
                            {
                                numTopics = e.topics.Length;
                                Console.WriteLine("num topics=" + numTopics);
                            } else
                            {
                                // guarantee num topics is consistent across the entire dataset
                                Debug.Assert(numTopics == e.topics.Length);
                            }
                        }

                        // google hits features
                        e.google_hits = new double[2];
                        e.google_hits[0] = Int32.Parse(eventFeatures[8]);
                        e.google_hits[1] = Int32.Parse(eventFeatures[9]);

                        // facebook likes features
                        e.facebook_likes = new double[3];
                        e.facebook_likes[0] = Int32.Parse(eventFeatures[10]);
                        e.facebook_likes[1] = Int32.Parse(eventFeatures[11]);
                        e.facebook_likes[2] = Int32.Parse(eventFeatures[12]);

                        events.Add(e);
                    }
                    xEvents.Add(events);
                    titlesList.Add(titlesSubList.ToArray());

                    readItems = readItems + 1;
                    currentLocation = currentLocation + 1;
                }
            }
            Console.WriteLine("unique venues found: " + venuesDict.Count);

            int numBaseFeatures = 1 + 9; // bias feature + rest
            int numEventsFeatures = 1 + 16;
            if (useVenuesInsteadOfStops)
            {
                numBaseFeatures += venuesDict.Count;
                numEventsFeatures += venuesDict.Count;
            }
            else
            {
                numBaseFeatures += stopsDict.Count;
                numEventsFeatures += stopsDict.Count;
            }
            if (useCategoriesAsFeatures)
            {
                Console.WriteLine("unique categories found:");
                foreach (string cat in categoriesDict.Keys) {
                    Console.WriteLine(cat);
                }
                numEventsFeatures += categoriesDict.Count;
            }
            if (useTopicsAsFeatures)
            {
                numEventsFeatures += numTopics;
            }
            List<Vector> baseVectors = new List<Vector>();
            List<Vector[]> eventsVectors = new List<Vector[]>();
            for (int i = 0; i < xEvents.Count; i++)
            {
                Event dummy = xBase[i];
                double[] baseFeaturesArray = new double[numBaseFeatures];
                Array.Clear(baseFeaturesArray, 0, baseFeaturesArray.Length);
                baseFeaturesArray[0] = 1; // bias feature
                int offset = 1;

                // stop_id
                baseFeaturesArray[offset + dummy.stop_id] = 1;
                offset += stopsDict.Count;
                
                // dow
                baseFeaturesArray[offset + dummy.dow - 1] = 1;
                offset += 7;

                // holiday
                baseFeaturesArray[offset] = dummy.holiday;
                offset += 1;

                // peak duration
                //baseFeaturesArray[offset] = dummy.end_time - dummy.start_time;
                //offset += 1;

                // stop mean
                baseFeaturesArray[offset] = dummy.expected_mean;

                Vector baseVector = Vector.FromArray(baseFeaturesArray);
                baseVectors.Add(baseVector);

                List<Vector> eventsList = new List<Vector>();
                for (int j = 0; j < xEvents[i].Count; j++)
                {
                    Event e = xEvents[i][j];
                    double[] eventFeatures = new double[numEventsFeatures];
                    Array.Clear(eventFeatures, 0, eventFeatures.Length);
                    eventFeatures[0] = 1; // bias feature
                    offset = 1;

                    // stop_id
                    if (useVenuesInsteadOfStops)
                    {
                        // venue
                        eventFeatures[offset + e.venue] = 1;
                        offset += venuesDict.Count;
                    }
                    else
                    {
                        // stop_id
                        eventFeatures[offset + e.stop_id] = 1;
                        offset += stopsDict.Count;
                    }

                    // dow
                    eventFeatures[offset + e.dow - 1] = 1;
                    offset += 7;

                    // holiday
                    eventFeatures[offset] = e.holiday;
                    offset += 1;

                    // peak start_time
                    /*if (e.start_time > 8 && e.start_time <= 24)
                    {
                        eventFeatures[offset] = 1;
                    }
                    else if (e.start_time > 24 && e.start_time <= 36)
                    {
                        eventFeatures[offset + 1] = 1;
                    }
                    else
                    {
                        eventFeatures[offset + 2] = 1;
                    }
                    offset += 3;
                    */

                    // stop mean, distance and all_day
                    eventFeatures[offset] = e.expected_mean;
                    //eventFeatures[offset + 1] = e.distance;
                    eventFeatures[offset + 1] = e.all_day;
                    eventFeatures[offset + 2] = e.repeated_count;
                    offset += 3;

                    if (useCategoriesAsFeatures)
                    {
                        double[] eventCategoriesBOW = new double[categoriesDict.Count];
                        Array.Clear(eventCategoriesBOW, 0, eventCategoriesBOW.Length);
                        int[] eventCategoriesInt = e.categories;
                        for (int k = 0; k < eventCategoriesInt.Length; k++)
                        {
                            eventCategoriesBOW[eventCategoriesInt[k]] += 1;
                        }
                        for (int k = 0; k < eventCategoriesBOW.Length; k++)
                        {
                            eventFeatures[offset + k] = eventCategoriesBOW[k];
                        }
                        offset += eventCategoriesBOW.Length;
                    }

                    if (useTopicsAsFeatures)
                    {
                        for (int k = 0; k < e.topics.Length; k++)
                        {
                            eventFeatures[offset + k] = e.topics[k];
                        }
                        offset += e.topics.Length;
                    }

                    for (int k = 0; k < e.google_hits.Length; k++)
                    {
                        eventFeatures[offset + k] = e.google_hits[k];
                    }
                    offset += e.google_hits.Length;

                    for (int k = 0; k < e.facebook_likes.Length; k++)
                    {
                        eventFeatures[offset + k] = e.facebook_likes[k];
                    }
                    offset += e.facebook_likes.Length;

                    Vector eventVector = Vector.FromArray(eventFeatures);
                    eventsList.Add(eventVector);
                }
                eventsVectors.Add(eventsList.ToArray());
            }

            titles = titlesList.ToArray();
            dates = datesList.ToArray();
            baseFeatures = baseVectors.ToArray();
            targets = targetsList.ToArray();
            return eventsVectors.ToArray();
        }

		private static double[][] Sparsify(Vector[] featureVectors, out int[][] indices)
		{
			int n = featureVectors.Length;
			double[][] values = new double[n][];
			indices = new int[n][];

			for (int i = 0; i < n; i++)
			{
				List<double> featureValues = new List<double>();
				List<int> featureIndices = new List<int>();
				for (int f = 0; f < featureVectors[i].Count; f++)
				{
					if (featureVectors[i][f] != valueToIgnore)
					{
						featureValues.Add(featureVectors[i][f]);
						featureIndices.Add(f);
					}
				}
				values[i] = featureValues.ToArray();
				indices[i] = featureIndices.ToArray();
			}
			return values;
		}

        public static Vector[] NumericToNominal(Vector[] data, int[] indices)
        {
            Vector[] convertedData = new Vector[data.Length];
            // initilize dictionaries and IDs
            Dictionary<double, int>[] maps = new Dictionary<double, int>[indices.Length];
            int[] nextIDs = new int[indices.Length];
            for (int j = 0; j < indices.Length; j++) 
            {
                maps[j] = new Dictionary<double,int>();
                nextIDs[j] = 0;
            }

            // first pass on the data: build dictionaries
            for (int i = 0; i < data.Length; i++)
            {
                Vector vector = (Vector) data.GetValue(i);
                for (int j = 0; j < indices.Length; j++) 
                {
                    double value = (double) vector.ToArray().GetValue(j);
                    if (!maps[j].ContainsKey(value))
                    {
                        maps[j].Add(value, nextIDs[j]);
                        nextIDs[j]++;
                    }
                }
            }

            // compute new feature vector size
            int newVectorSize = 0;
            for (int j=0; j<indices.Length; j++) 
            {
                newVectorSize += maps[j].Count;
            }

            // second pass on the data: generate new vectors
            for (int i = 0; i < data.Length; i++)
            {
                Vector vector = (Vector) data.GetValue(i);

                // initilize new vector with zeros
                double[] newVector = new double[newVectorSize];
                for (int j = 0; j < newVectorSize; j++)
                {
                    newVector[j] = 0;
                }

                // make conversion
                int offset = 0;
                for (int j = 0; j < indices.Length; j++)
                {
                    double value = (double) vector.ToArray().GetValue(j);
                    int mappedValue;
                    maps[j].TryGetValue(value, out mappedValue);
                    newVector[offset + mappedValue] = 1;
                    offset += maps[j].Count;
                }

                convertedData[i] = Vector.FromArray(newVector);
            }

            return convertedData;
        }
	}

    public class Event
    {
        public int[] categories;
        public double[] topics;
        public double distance;
        public string id;
        public int stop_id;
        public int dow;
        public int holiday;
        public int start_time;
        public int end_time;
        public int repeated_count;
        public int venue;
        public double area;
        public double rmse;
        public double overall_mean;
        public double expected_mean;
        public double observed_mean;
        public int all_day;
        public double[] google_hits;
        public double[] facebook_likes;
    }
}