﻿/********************************************************
*                                                       *
*   Copyright (C) Microsoft. All rights reserved.       *
*                                                       *
********************************************************/

namespace EventsModel
{
	using System;
	using System.Collections.Generic;

	using MicrosoftResearch.Infer;
	using MicrosoftResearch.Infer.Distributions;
	using MicrosoftResearch.Infer.Maths;
	using MicrosoftResearch.Infer.Utils;
    using System.Globalization;
    using MicrosoftResearch.Infer.Models;
    using System.Diagnostics;

	class Program
    {
        /// <summary>
		/// The entry point for the program.
		/// </summary>
		/// <param name="args">The array of command-line arguments.</param>
		static void Main(string[] args)
        {
            Console.WriteLine("Starting...");
            System.Globalization.CultureInfo customCulture = (System.Globalization.CultureInfo)System.Threading.Thread.CurrentThread.CurrentCulture.Clone();
            customCulture.NumberFormat.NumberDecimalSeparator = ".";
            System.Threading.Thread.CurrentThread.CurrentCulture = customCulture;

            // load data
			Console.WriteLine("Loading data...");
            bool ignoreHeader;
            string datadir = @"..\..\data\"; //windows
            //string datadir = @"../../data/"; //unix
            string datafile = datadir + @"63.tsv"; 
            Console.WriteLine("Dataset: " + datafile);
            int[] featuresToIgnore = new int[] { 0 };
            string[] dates;
            string[][] titles;
            double[] targets;
            Vector[] baseVectors;
            Vector[][] eventsVectors = DataUtils.Read(datafile, ignoreHeader = true, out baseVectors, out targets, out dates, out titles);

            // Build dataset
            DataSet dataset = new DataSet(baseVectors, eventsVectors, targets, dates, titles);
            //dataset = StandardizeDataset(dataset);
            dataset = ShuffleDataset(dataset);
            DataSet trainset;
            DataSet testset;
            //TrainTestSplit(dataset, 0.9, out trainset, out testset);
            int fold = 9; // ranges from 0 to 9
            FoldSplit(dataset, fold, out trainset, out testset);
            Console.WriteLine("num instances=" + dataset.GetNumInstances());
            Console.WriteLine("num base features=" + dataset.GetNumBaseFeatures());
            Console.WriteLine("num events features=" + dataset.GetNumEventsFeatures());

            //Console.WriteLine("converted features=" + StringUtil.ArrayToString(baseVectors));
            System.IO.File.WriteAllText(datadir + @"baseFeaturesDebug.txt", StringUtil.ArrayToString(baseVectors));
            //Console.WriteLine("events features=" + StringUtil.ArrayToString(eventsFeatureVectors));
            System.IO.File.WriteAllText(datadir + @"eventsFeaturesDebug.txt", StringUtil.ArrayToString(eventsVectors));

            EventsLRModel eventsModel = new EventsLRModel();
            Console.WriteLine("computing model evidence...");
            eventsModel.CreateModelTraining(dataset.GetNumBaseFeatures(), dataset.GetNumEventsFeatures());
            double logEvidence = eventsModel.ComputeModelEvidence(dataset);
            Console.WriteLine("log evidence=" + logEvidence);
            Console.WriteLine("estimating model parameters...");
            EventsModelData weightsPosteriors = eventsModel.InferModelWeights(dataset);
            System.IO.File.WriteAllText(datadir + @"eventsWeightsDebug.txt", weightsPosteriors.eventsWeightsPosterior.ToString());
            System.IO.File.WriteAllText(datadir + @"dummyWeightsDebug.txt", weightsPosteriors.dummyWeightsPosterior.ToString());
            eventsModel.CreateModelTesting(dataset.GetNumBaseFeatures(), dataset.GetNumEventsFeatures());
            
            Console.WriteLine("making predictions...");
            eventsModel.CreateModelTesting(testset.GetNumBaseFeatures(), testset.GetNumEventsFeatures());
            DistributionStructArray<Gaussian, double> predictions = eventsModel.MakePredictions(weightsPosteriors, testset);
            double[] preds = ComputeError(testset.targets, predictions);
            using (System.IO.StreamWriter file = new System.IO.StreamWriter(datafile.Replace(".tsv", "_fold" + fold + ".csv")))
            {
                file.WriteLine("date,time,true,pred");
                for (int i = 0; i < testset.targets.Length; i++)
                {
                    file.WriteLine(testset.dates[i] + "," + testset.targets[i] + "," + preds[i]);
                }
                file.Flush();
                file.Close();
            }

			Console.WriteLine("Press key to quit");
			Console.ReadKey();
		}

        public static double[] ComputeError(double[] targets, DistributionStructArray<Gaussian, double> yPosterior)
        {
            Array predictiveDists = yPosterior.ToArray();
            double[] preds = new double[targets.Length];
            for (int i = 0; i < targets.Length; i++)
            {
                Gaussian predictiveDist = (Gaussian)predictiveDists.GetValue(i);
                //Console.WriteLine(targets[i] + " - " + predictiveDist);
                preds[i] = predictiveDist.GetMean();
            }
            return ComputeError(targets, preds);
        }

        public static double[] ComputeError(double[] targets, double[] preds)
        {

            double meanTarget = MathNet.Numerics.Statistics.ArrayStatistics.Mean(targets);
            double sumError = 0;
            double sumSquaredError = 0;
            double sumNorms = 0;
            double sumNormsSquared = 0;
            for (int i = 0; i < targets.Length; i++)
            {
                double error = targets[i] - preds[i];
                sumError += Math.Abs(error);
                sumSquaredError += Math.Pow(error, 2);
                sumNorms += Math.Abs(targets[i] - meanTarget);
                sumNormsSquared += Math.Pow(targets[i] - meanTarget, 2);
            }
            double corr = MathNet.Numerics.Statistics.Correlation.Pearson(targets, preds);
            double mae = sumError / targets.Length;
            double rmse = Math.Sqrt(sumSquaredError / targets.Length);
            double rae = sumError / sumNorms;
            double rrse = sumSquaredError / sumNormsSquared;
            Console.WriteLine("\nGLOBAL ERROR STATISTICS:");
            Console.WriteLine("corr=" + (corr * 100).ToString("0.000") + "\tmae=" + mae.ToString("0.000") + "\trmse=" + rmse.ToString("0.000") + "\trae=" + (rae * 100).ToString("0.000") + "\trrse=" + (rrse * 100).ToString("0.000"));
            Console.WriteLine("TSV:" + (corr * 100).ToString("0.000") + "\t" + mae.ToString("0.000") + "\t" + rmse.ToString("0.000") + "\t" + (rae * 100).ToString("0.000") + "\t" + (rrse * 100).ToString("0.000"));
            return preds;
        }

        public static DataSet ShuffleDataset(DataSet dataset)
        {
            int seed = 0;
            Shuffle<Vector>(new Random(seed), dataset.baseFeatureVectors);
            Shuffle<Vector[]>(new Random(seed), dataset.eventsFeatureVectors);
            Shuffle<double>(new Random(seed), dataset.targets);
            Shuffle<string>(new Random(seed), dataset.dates);
            Shuffle<string[]>(new Random(seed), dataset.titles);
            return dataset;
        }

        public static void Shuffle<T>(Random rng, T[] array)
        {
            int n = array.Length;
            while (n > 1)
            {
                int k = rng.Next(n--);
                T temp = array[n];
                array[n] = array[k];
                array[k] = temp;
            }
        }

        public static void TrainTestSplit(DataSet dataset, double percentage, out DataSet trainset, out DataSet testset)
        {
            trainset = new DataSet();
            testset = new DataSet();

            int splitPoint = (int)(dataset.GetNumInstances() * percentage);
            SplitArray<Vector>(dataset.baseFeatureVectors, splitPoint, out trainset.baseFeatureVectors, out testset.baseFeatureVectors);
            SplitArray<Vector[]>(dataset.eventsFeatureVectors, splitPoint, out trainset.eventsFeatureVectors, out testset.eventsFeatureVectors);
            SplitArray<double>(dataset.targets, splitPoint, out trainset.targets, out testset.targets);
            SplitArray<string>(dataset.dates, splitPoint, out trainset.dates, out testset.dates);
            SplitArray<string[]>(dataset.titles, splitPoint, out trainset.titles, out testset.titles);
        }

        public static void FoldSplit(DataSet dataset, int fold, out DataSet trainset, out DataSet testset)
        {
            trainset = new DataSet();
            testset = new DataSet();

            int splitPoint1 = (int)(dataset.GetNumInstances() * (fold / 10.0));
            int splitPoint2 = (int)(dataset.GetNumInstances() * ((fold + 1) / 10.0));
            Console.WriteLine("splitPoint1=" + splitPoint1 + " splitPoint2=" + splitPoint2);
            SplitArray<Vector>(dataset.baseFeatureVectors, splitPoint1, splitPoint2, out trainset.baseFeatureVectors, out testset.baseFeatureVectors);
            SplitArray<Vector[]>(dataset.eventsFeatureVectors, splitPoint1, splitPoint2, out trainset.eventsFeatureVectors, out testset.eventsFeatureVectors);
            SplitArray<double>(dataset.targets, splitPoint1, splitPoint2, out trainset.targets, out testset.targets);
            SplitArray<string>(dataset.dates, splitPoint1, splitPoint2, out trainset.dates, out testset.dates);
            SplitArray<string[]>(dataset.titles, splitPoint1, splitPoint2, out trainset.titles, out testset.titles);
        }

        public static void SplitArray<T>(T[] source, int index, out T[] first, out T[] second)
        {
            int len2 = source.Length - index;
            first = new T[index];
            second = new T[len2];
            Array.Copy(source, 0, first, 0, index);
            Array.Copy(source, index, second, 0, len2);
        }

        public static void SplitArray<T>(T[] source, int index1, int index2, out T[] first, out T[] second)
        {
            Debug.Assert(index2 > index1);
            first = new T[source.Length - (index2 - index1)];
            second = new T[index2 - index1];
            Array.Copy(source, 0, first, 0, index1);
            Array.Copy(source, index2, first, index1, source.Length - index2);
            Array.Copy(source, index1, second, 0, index2 - index1);
        }
	}
}